@@ -51,6 +51,9 @@
 #ifndef S_ISSOCK
 #define S_ISSOCK(x) (0)
 #endif
+#ifndef O_BINARY
+#define O_BINARY (0)
+#endif
 
 /**********************************************************************
  * GStreamer Default File Source
@@ -756,7 +759,7 @@ gst_filesrc_open_file (GstFileSrc * src)
   GST_INFO_OBJECT (src, "opening file %s", src->filename);
 
   /* open the file */
-  src->fd = open (src->filename, O_RDONLY);
+  src->fd = open (src->filename, O_RDONLY | O_BINARY);
   if (src->fd < 0) {
     if (errno == ENOENT)
       GST_ELEMENT_ERROR (src, RESOURCE, NOT_FOUND, (NULL),
@@ -912,7 +915,7 @@ gst_filesrc_srcpad_event (GstPad * pad, GstEvent * event)
 {
   GstFileSrc *src = GST_FILESRC (GST_PAD_PARENT (pad));
 
-  GST_DEBUG_OBJECT (src, "event %d", GST_EVENT_TYPE (event));
+  GST_DEBUG_OBJECT (src, "received event type %d", GST_EVENT_TYPE (event));
 
   switch (GST_EVENT_TYPE (event)) {
     case GST_EVENT_SEEK:
