@@ -32,6 +32,7 @@
 
 #include "gstscheduler.h"
 #include "gstindex.h"
+#include "gstutils.h"
 
 static GstElementDetails gst_bin_details = GST_ELEMENT_DETAILS ("Generic bin",
     "Generic/Bin",
@@ -56,6 +57,8 @@ static void gst_bin_add_func (GstBin * bin, GstElement * element);
 static void gst_bin_remove_func (GstBin * bin, GstElement * element);
 static void gst_bin_child_state_change_func (GstBin * bin,
     GstElementState oldstate, GstElementState newstate, GstElement * child);
+GstElementStateReturn gst_bin_set_state (GstElement * element,
+    GstElementState state);
 
 static GstClock *gst_bin_get_clock_func (GstElement * element);
 static void gst_bin_set_clock_func (GstElement * element, GstClock * clock);
@@ -161,6 +164,7 @@ gst_bin_class_init (GstBinClass * klass)
 #endif
 
   gstelement_class->change_state = GST_DEBUG_FUNCPTR (gst_bin_change_state);
+  gstelement_class->set_state = GST_DEBUG_FUNCPTR (gst_bin_set_state);
 #ifndef GST_DISABLE_INDEX
   gstelement_class->set_index = GST_DEBUG_FUNCPTR (gst_bin_set_index);
 #endif
@@ -455,6 +459,12 @@ gst_bin_add_func (GstBin * bin, GstElement * element)
     return;
   }
 
+  if (GST_STATE (element) > GST_STATE (bin)) {
+    GST_CAT_DEBUG_OBJECT (GST_CAT_STATES, bin,
+        "setting state to receive element \"%s\"", GST_OBJECT_NAME (element));
+    gst_element_set_state ((GstElement *) bin, GST_STATE (element));
+  }
+
   /* set the element's parent and add the element to the bin's list of children */
   gst_object_set_parent (GST_OBJECT (element), GST_OBJECT (bin));
 
@@ -588,7 +598,6 @@ gst_bin_remove (GstBin * bin, GstElement * element)
 
   g_return_if_fail (GST_IS_BIN (bin));
   g_return_if_fail (GST_IS_ELEMENT (element));
-  g_return_if_fail (bin->children != NULL);
 
   bclass = GST_BIN_GET_CLASS (bin);
 
@@ -665,11 +674,14 @@ static void
 gst_bin_child_state_change_func (GstBin * bin, GstElementState oldstate,
     GstElementState newstate, GstElement * child)
 {
+  GstElementState old = 0, new = 0;
   gint old_idx = 0, new_idx = 0, i;
 
-  while (oldstate >>= 1)
+  old = oldstate;
+  new = newstate;
+  while (old >>= 1)
     old_idx++;
-  while (newstate >>= 1)
+  while (new >>= 1)
     new_idx++;
 
   GST_LOCK (bin);
@@ -699,16 +711,110 @@ gst_bin_child_state_change_func (GstBin * bin, GstElementState oldstate,
   GST_UNLOCK (bin);
 }
 
+typedef gboolean (*GstBinForeachFunc) (GstBin * bin, GstElement * element,
+    gpointer data);
+
+/*
+ * gst_bin_foreach:
+ * @bin: bin to traverse
+ * @func: function to call on each child
+ * @data: user data handed to each function call
+ *
+ * Calls @func on each child of the bin. If @func returns FALSE, 
+ * gst_bin_foreach() immediately returns.
+ * It is assumed that calling @func may alter the set of @bin's children. @func
+ * will only be called on children that were in @bin when gst_bin_foreach() was
+ * called, and that are still in @bin when the child is reached.
+ *
+ * Returns: TRUE if @func always returned TRUE, FALSE otherwise
+ **/
+static gboolean
+gst_bin_foreach (GstBin * bin, GstBinForeachFunc func, gpointer data)
+{
+  GList *kids, *walk;
+
+  g_return_val_if_fail (GST_IS_BIN (bin), FALSE);
+  g_return_val_if_fail (func != NULL, FALSE);
+
+  kids = g_list_copy (bin->children);
+
+  for (walk = kids; walk; walk = g_list_next (walk)) {
+    GstElement *element = (GstElement *) walk->data;
+
+    if (g_list_find (bin->children, element)) {
+      gboolean res = func (bin, element, data);
+
+      if (!res) {
+        g_list_free (kids);
+        return FALSE;
+      }
+    }
+  }
+
+  g_list_free (kids);
+  return TRUE;
+}
+
+typedef struct
+{
+  GstElementState pending;
+  GstElementStateReturn result;
+}
+SetKidStateData;
+static int
+set_kid_state_func (GstBin * bin, GstElement * child, gpointer user_data)
+{
+  GstElementState old_child_state;
+  SetKidStateData *data = user_data;
+
+  if (GST_FLAG_IS_SET (child, GST_ELEMENT_LOCKED_STATE)) {
+    return TRUE;
+  }
+
+  old_child_state = GST_STATE (child);
+
+  GST_CAT_DEBUG_OBJECT (GST_CAT_STATES, bin,
+      "changing state of child %s from current %s to pending %s",
+      GST_ELEMENT_NAME (child), gst_element_state_get_name (old_child_state),
+      gst_element_state_get_name (data->pending));
+
+  switch (gst_element_set_state (child, data->pending)) {
+    case GST_STATE_FAILURE:
+      GST_CAT_INFO_OBJECT (GST_CAT_STATES, bin,
+          "child '%s' failed to go to state %d(%s)",
+          GST_ELEMENT_NAME (child),
+          data->pending, gst_element_state_get_name (data->pending));
+
+      gst_element_set_state (child, old_child_state);
+      return FALSE;             /* error out to the caller */
+
+    case GST_STATE_ASYNC:
+      GST_CAT_INFO_OBJECT (GST_CAT_STATES, bin,
+          "child '%s' is changing state asynchronously",
+          GST_ELEMENT_NAME (child));
+      data->result = GST_STATE_ASYNC;
+      return TRUE;
+
+    case GST_STATE_SUCCESS:
+      GST_CAT_DEBUG (GST_CAT_STATES,
+          "child '%s' changed state to %d(%s) successfully",
+          GST_ELEMENT_NAME (child), data->pending,
+          gst_element_state_get_name (data->pending));
+      return TRUE;
+
+    default:
+      g_assert_not_reached ();
+      return FALSE;             /* satisfy gcc */
+  }
+}
+
 static GstElementStateReturn
 gst_bin_change_state (GstElement * element)
 {
   GstBin *bin;
-  GList *children;
-  GstElement *child;
   GstElementStateReturn ret;
   GstElementState old_state, pending;
-  gint transition;
-  gboolean have_async = FALSE;
+  SetKidStateData data;
 
   g_return_val_if_fail (GST_IS_BIN (element), GST_STATE_FAILURE);
 
@@ -716,7 +822,6 @@ gst_bin_change_state (GstElement * element)
 
   old_state = GST_STATE (element);
   pending = GST_STATE_PENDING (element);
-  transition = GST_STATE_TRANSITION (element);
 
   GST_CAT_DEBUG_OBJECT (GST_CAT_STATES, element,
       "changing state of children from %s to %s",
@@ -726,65 +831,11 @@ gst_bin_change_state (GstElement * element)
   if (pending == GST_STATE_VOID_PENDING)
     return GST_STATE_SUCCESS;
 
-  if (old_state == pending) {
-    GST_CAT_LOG_OBJECT (GST_CAT_STATES, element,
-        "old and pending state are both %s, returning",
-        gst_element_state_get_name (pending));
-    return GST_STATE_SUCCESS;
-  }
-
-  children = bin->children;
-
-  while (children) {
-    GstElementState old_child_state;
-
-    child = GST_ELEMENT (children->data);
-
-    if (GST_FLAG_IS_SET (child, GST_ELEMENT_LOCKED_STATE)) {
-      children = g_list_next (children);
-      continue;
-    }
-
-    old_child_state = GST_STATE (child);
-
-    GST_CAT_DEBUG_OBJECT (GST_CAT_STATES, element,
-        "changing state of child %s from current %s to pending %s",
-        GST_ELEMENT_NAME (child), gst_element_state_get_name (old_child_state),
-        gst_element_state_get_name (pending));
-
-    switch (gst_element_set_state (child, pending)) {
-      case GST_STATE_FAILURE:
-        GST_CAT_INFO_OBJECT (GST_CAT_STATES, element,
-            "child '%s' failed to go to state %d(%s)",
-            GST_ELEMENT_NAME (child),
-            pending, gst_element_state_get_name (pending));
-
-        gst_element_set_state (child, old_child_state);
-        /* There was a check for elements being in the same scheduling group
-           here. Removed by dolphy <julien@moutte.net>. No matter the 
-           scheduling group we should always return a failure. This change
-           seems to work on my machine and fixes tons of issues. If anyone
-           want to revert please tell me what it breaks first, Thanks. */
-        GST_STATE_PENDING (element) = old_state;
-        return GST_STATE_FAILURE;
-        break;
-      case GST_STATE_ASYNC:
-        GST_CAT_INFO_OBJECT (GST_CAT_STATES, element,
-            "child '%s' is changing state asynchronously",
-            GST_ELEMENT_NAME (child));
-        have_async = TRUE;
-        break;
-      case GST_STATE_SUCCESS:
-        GST_CAT_DEBUG (GST_CAT_STATES,
-            "child '%s' changed state to %d(%s) successfully",
-            GST_ELEMENT_NAME (child), pending,
-            gst_element_state_get_name (pending));
-        break;
-    }
-    /* we need to do this down here, because there might be elements removed 
-     * from this bin during state changes, so g_list_next (children) might
-     * change as well */
-    children = g_list_next (children);
+  data.pending = pending;
+  data.result = GST_STATE_SUCCESS;
+  if (!gst_bin_foreach (bin, set_kid_state_func, &data)) {
+    GST_STATE_PENDING (element) = old_state;
+    return GST_STATE_FAILURE;
   }
 
   GST_CAT_DEBUG_OBJECT (GST_CAT_STATES, element,
@@ -793,9 +844,10 @@ gst_bin_change_state (GstElement * element)
       gst_element_state_get_name (pending),
       gst_element_state_get_name (GST_STATE (element)));
 
-  if (have_async)
+  if (data.result == GST_STATE_ASYNC)
     ret = GST_STATE_ASYNC;
   else {
+    /* FIXME: this should have been done by the children already, no? */
     if (parent_class->change_state) {
       ret = parent_class->change_state (element);
     } else
@@ -804,6 +856,26 @@ gst_bin_change_state (GstElement * element)
   return ret;
 }
 
+GstElementStateReturn
+gst_bin_set_state (GstElement * element, GstElementState state)
+{
+  GstBin *bin = GST_BIN (element);
+
+  if (GST_STATE (bin) == state) {
+    SetKidStateData data;
+
+    data.pending = state;
+    data.result = GST_STATE_SUCCESS;
+    if (!gst_bin_foreach (bin, set_kid_state_func, &data)) {
+      return GST_STATE_FAILURE;
+    } else {
+      return data.result;
+    }
+  } else {
+    return GST_CALL_PARENT_WITH_DEFAULT (GST_ELEMENT_CLASS, set_state, (element,
+            state), GST_STATE_FAILURE);
+  }
+}
 
 static GstElementStateReturn
 gst_bin_change_state_norecurse (GstBin * bin)
@@ -826,8 +898,7 @@ gst_bin_dispose (GObject * object)
 
   GST_CAT_DEBUG_OBJECT (GST_CAT_REFCOUNTING, object, "dispose");
 
-  if (gst_element_get_state (GST_ELEMENT (object)) == GST_STATE_PLAYING)
-    gst_element_set_state (GST_ELEMENT (object), GST_STATE_PAUSED);
+  gst_element_set_state (GST_ELEMENT (object), GST_STATE_NULL);
 
   while (bin->children) {
     gst_bin_remove (bin, GST_ELEMENT (bin->children->data));
@@ -931,10 +1002,10 @@ gst_bin_get_list (GstBin * bin)
  * @bin: bin to find element in
  * @interface: interface to be implemented by interface
  *
- * Looks for the first element inside the bin that implements the given 
+ * Looks for the first element inside the bin that implements the given
  * interface. If such an element is found, it returns the element. You can
  * cast this element to the given interface afterwards.
- * If you want all elements that implement the interface, use 
+ * If you want all elements that implement the interface, use
  * gst_bin_get_all_by_interface(). The function recurses bins inside bins.
  *
  * Returns: An element inside the bin implementing the interface.
@@ -969,9 +1040,9 @@ gst_bin_get_by_interface (GstBin * bin, GType interface)
  * @bin: bin to find elements in
  * @interface: interface to be implemented by interface
  *
- * Looks for all element inside the bin that implements the given 
+ * Looks for all elements inside the bin that implements the given
  * interface. You can safely cast all returned elements to the given interface.
- * The function recurses bins inside bins. You need to free the list using 
+ * The function recurses bins inside bins. You need to free the list using
  * g_list_free() after use.
  *
  * Returns: An element inside the bin implementing the interface.
@@ -986,8 +1057,11 @@ gst_bin_get_all_by_interface (GstBin * bin, GType interface)
 
   walk = bin->children;
   while (walk) {
-    if (G_TYPE_CHECK_INSTANCE_TYPE (walk->data, interface))
+    if (G_TYPE_CHECK_INSTANCE_TYPE (walk->data, interface)) {
+      GST_DEBUG_OBJECT (bin, "element %s implements requested interface",
+          GST_ELEMENT_NAME (GST_ELEMENT (walk->data)));
       ret = g_list_prepend (ret, walk->data);
+    }
     if (GST_IS_BIN (walk->data)) {
       ret = g_list_concat (ret,
           gst_bin_get_all_by_interface (GST_BIN (walk->data), interface));
@@ -1106,6 +1180,8 @@ gst_bin_restore_thyself (GstObject * object, xmlNodePtr self)
 
     field = field->next;
   }
+  if (GST_OBJECT_CLASS (parent_class)->restore_thyself)
+    (GST_OBJECT_CLASS (parent_class)->restore_thyself) (object, self);
 }
 #endif /* GST_DISABLE_LOADSAVE */
 
