@@ -26,20 +26,23 @@
 #include "gstelement.h"
 #include "gstregistrypool.h"
 #include "gstinfo.h"
+#include "gsturi.h"
+#include "registries/gstxmlregistry.h"  /* g_critical in gst_element_factory_create */
 
 GST_DEBUG_CATEGORY_STATIC (element_factory_debug);
 #define GST_CAT_DEFAULT element_factory_debug
 
-static void 		gst_element_factory_class_init 		(GstElementFactoryClass *klass);
-static void 		gst_element_factory_init 		(GstElementFactory *factory);
+static void gst_element_factory_class_init (GstElementFactoryClass * klass);
+static void gst_element_factory_init (GstElementFactory * factory);
 
-static void 		gst_element_factory_unload_thyself 	(GstPluginFeature *feature);
+static void gst_element_factory_unload_thyself (GstPluginFeature * feature);
 
 static GstPluginFeatureClass *parent_class = NULL;
+
 /* static guint gst_element_factory_signals[LAST_SIGNAL] = { 0 }; */
 
-GType 
-gst_element_factory_get_type (void) 
+GType
+gst_element_factory_get_type (void)
 {
   static GType elementfactory_type = 0;
 
@@ -51,40 +54,48 @@ gst_element_factory_get_type (void)
       (GClassInitFunc) gst_element_factory_class_init,
       NULL,
       NULL,
-      sizeof(GstElementFactory),
+      sizeof (GstElementFactory),
       0,
       (GInstanceInitFunc) gst_element_factory_init,
       NULL
     };
-    elementfactory_type = g_type_register_static (GST_TYPE_PLUGIN_FEATURE, 
-		    				  "GstElementFactory", &elementfactory_info, 0);
-    GST_DEBUG_CATEGORY_INIT (element_factory_debug, "GST_ELEMENT_FACTORY", 
-	    GST_DEBUG_BOLD | GST_DEBUG_FG_WHITE | GST_DEBUG_BG_RED,
-	    "element factories keep information about installed elements");
+
+    elementfactory_type = g_type_register_static (GST_TYPE_PLUGIN_FEATURE,
+        "GstElementFactory", &elementfactory_info, 0);
+    GST_DEBUG_CATEGORY_INIT (element_factory_debug, "GST_ELEMENT_FACTORY",
+        GST_DEBUG_BOLD | GST_DEBUG_FG_WHITE | GST_DEBUG_BG_RED,
+        "element factories keep information about installed elements");
   }
   return elementfactory_type;
 }
 static void
-gst_element_factory_class_init (GstElementFactoryClass *klass)
+gst_element_factory_class_init (GstElementFactoryClass * klass)
 {
   GObjectClass *gobject_class;
   GstObjectClass *gstobject_class;
   GstPluginFeatureClass *gstpluginfeature_class;
 
-  gobject_class = (GObjectClass*)klass;
-  gstobject_class = (GstObjectClass*)klass;
-  gstpluginfeature_class = (GstPluginFeatureClass*) klass;
+  gobject_class = (GObjectClass *) klass;
+  gstobject_class = (GstObjectClass *) klass;
+  gstpluginfeature_class = (GstPluginFeatureClass *) klass;
 
   parent_class = g_type_class_peek_parent (klass);
 
-  gstpluginfeature_class->unload_thyself = 	GST_DEBUG_FUNCPTR (gst_element_factory_unload_thyself);
+  gstpluginfeature_class->unload_thyself =
+      GST_DEBUG_FUNCPTR (gst_element_factory_unload_thyself);
 }
 static void
-gst_element_factory_init (GstElementFactory *factory)
+gst_element_factory_init (GstElementFactory * factory)
 {
   factory->padtemplates = NULL;
   factory->numpadtemplates = 0;
+
+  factory->uri_type = GST_URI_UNKNOWN;
+  factory->uri_protocols = NULL;
+
+  factory->interfaces = NULL;
 }
+
 /**
  * gst_element_factory_find:
  * @name: name of factory to find
@@ -93,24 +104,25 @@ gst_element_factory_init (GstElementFactory *factory)
  *
  * Returns: #GstElementFactory if found, NULL otherwise
  */
-GstElementFactory*
-gst_element_factory_find (const gchar *name)
+GstElementFactory *
+gst_element_factory_find (const gchar * name)
 {
   GstPluginFeature *feature;
 
-  g_return_val_if_fail(name != NULL, NULL);
+  g_return_val_if_fail (name != NULL, NULL);
 
   feature = gst_registry_pool_find_feature (name, GST_TYPE_ELEMENT_FACTORY);
   if (feature)
     return GST_ELEMENT_FACTORY (feature);
 
-  /* this should be an ERROR */
-  GST_DEBUG ("no such elementfactory \"%s\"", name);
+  /* this isn't an error, for instance when you query if an element factory is
+   * present */
+  GST_LOG ("no such element factory \"%s\"", name);
   return NULL;
 }
 
 void
-__gst_element_details_clear (GstElementDetails *dp)
+__gst_element_details_clear (GstElementDetails * dp)
 {
   g_free (dp->longname);
   dp->longname = NULL;
@@ -121,22 +133,38 @@ __gst_element_details_clear (GstElementDetails *dp)
   g_free (dp->author);
   dp->author = NULL;
 }
+
+#define VALIDATE_SET(__dest, __src, __entry)				\
+G_STMT_START {								\
+  if (g_utf8_validate (__src->__entry, -1, NULL)) {			\
+    __dest->__entry = g_strdup (__src->__entry);			\
+  } else {								\
+    g_warning ("Invalid UTF-8 in " G_STRINGIFY (__entry) ": %s",	\
+        __src->__entry);						\
+    __dest->__entry = g_strdup ("[ERROR: invalid UTF-8]");		\
+  }									\
+} G_STMT_END
+
 void
-__gst_element_details_set (GstElementDetails *dest, const GstElementDetails *src)
+__gst_element_details_set (GstElementDetails * dest,
+    const GstElementDetails * src)
 {
-  dest->longname = g_strdup (src->longname);
-  dest->klass = g_strdup (src->klass);
-  dest->description = g_strdup (src->description);
-  dest->author = g_strdup (src->author);
+  VALIDATE_SET (dest, src, longname);
+  VALIDATE_SET (dest, src, klass);
+  VALIDATE_SET (dest, src, description);
+  VALIDATE_SET (dest, src, author);
 }
+
 void
-__gst_element_details_copy (GstElementDetails *dest, const GstElementDetails *src)
+__gst_element_details_copy (GstElementDetails * dest,
+    const GstElementDetails * src)
 {
   __gst_element_details_clear (dest);
   __gst_element_details_set (dest, src);
 }
+
 static void
-gst_element_factory_cleanup (GstElementFactory *factory)
+gst_element_factory_cleanup (GstElementFactory * factory)
 {
   __gst_element_details_clear (&factory->details);
   if (factory->type) {
@@ -144,27 +172,40 @@ gst_element_factory_cleanup (GstElementFactory *factory)
     factory->type = 0;
   }
 
-  g_list_foreach (factory->padtemplates, (GFunc) g_object_unref, NULL);
+  g_list_foreach (factory->padtemplates, (GFunc) gst_object_unref, NULL);
   g_list_free (factory->padtemplates);
   factory->padtemplates = NULL;
   factory->numpadtemplates = 0;
+  factory->uri_type = GST_URI_UNKNOWN;
+  if (factory->uri_protocols) {
+    g_strfreev (factory->uri_protocols);
+    factory->uri_protocols = NULL;
+  }
+
+  g_list_foreach (factory->interfaces, (GFunc) g_free, NULL);
+  g_list_free (factory->interfaces);
+  factory->interfaces = NULL;
 }
+
 /**
  * gst_element_register:
- * @plugin:
+ * @plugin: #GstPlugin to register the element with
  * @name: name of elements of this type
  * @rank: rank of element (higher rank means more importance when autoplugging)
  * @type: GType of element to register
  *
- * Create a new elementfactory capable of insantiating objects of the
+ * Create a new elementfactory capable of instantiating objects of the
  * given type.
  *
  * Returns: TRUE, if the registering succeeded, FALSE on error
  */
 gboolean
-gst_element_register (GstPlugin *plugin, const gchar *name, guint rank, GType type)
+gst_element_register (GstPlugin * plugin, const gchar * name, guint rank,
+    GType type)
 {
   GstElementFactory *factory;
+  GType *interfaces;
+  guint n_interfaces, i;
   GstElementClass *klass;
 
   g_return_val_if_fail (name != NULL, FALSE);
@@ -174,28 +215,56 @@ gst_element_register (GstPlugin *plugin, const gchar *name, guint rank, GType ty
 
   if (!factory) {
     klass = GST_ELEMENT_CLASS (g_type_class_ref (type));
-    factory = GST_ELEMENT_FACTORY (g_object_new (GST_TYPE_ELEMENT_FACTORY, NULL));
+    factory =
+        GST_ELEMENT_FACTORY (g_object_new (GST_TYPE_ELEMENT_FACTORY, NULL));
     gst_plugin_feature_set_name (GST_PLUGIN_FEATURE (factory), name);
-    GST_LOG_OBJECT (factory, "Created new elementfactory for type %s", g_type_name (type));
+    GST_LOG_OBJECT (factory, "Created new elementfactory for type %s",
+        g_type_name (type));
   } else {
     g_return_val_if_fail (factory->type == 0, FALSE);
     klass = GST_ELEMENT_CLASS (g_type_class_ref (type));
     gst_element_factory_cleanup (factory);
-    GST_LOG_OBJECT (factory, "Reuse existing elementfactory for type %s", g_type_name (type));
+    GST_LOG_OBJECT (factory, "Reuse existing elementfactory for type %s",
+        g_type_name (type));
   }
 
   factory->type = type;
   __gst_element_details_copy (&factory->details, &klass->details);
   factory->padtemplates = g_list_copy (klass->padtemplates);
-  g_list_foreach (factory->padtemplates, (GFunc) g_object_ref, NULL);
+  g_list_foreach (factory->padtemplates, (GFunc) gst_object_ref, NULL);
   factory->numpadtemplates = klass->numpadtemplates;
 
-  gst_plugin_feature_set_rank (GST_PLUGIN_FEATURE (factory), rank);
+  /* special stuff for URI handling */
+  if (g_type_is_a (type, GST_TYPE_URI_HANDLER)) {
+    GstURIHandlerInterface *iface = (GstURIHandlerInterface *)
+        g_type_interface_peek (klass, GST_TYPE_URI_HANDLER);
 
+    if (!iface || !iface->get_type || !iface->get_protocols)
+      goto error;
+    factory->uri_type = iface->get_type ();
+    if (!GST_URI_TYPE_IS_VALID (factory->uri_type))
+      goto error;
+    factory->uri_protocols = g_strdupv (iface->get_protocols ());
+    if (!factory->uri_protocols)
+      goto error;
+  }
+
+  interfaces = g_type_interfaces (type, &n_interfaces);
+  for (i = 0; i < n_interfaces; i++) {
+    __gst_element_factory_add_interface (factory, g_type_name (interfaces[i]));
+  }
+  g_free (interfaces);
+
+  gst_plugin_feature_set_rank (GST_PLUGIN_FEATURE (factory), rank);
   gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));
 
   return TRUE;
+
+error:
+  gst_element_factory_cleanup (factory);
+  return FALSE;
 }
+
 /**
  * gst_element_factory_create:
  * @factory: factory to instantiate
@@ -207,9 +276,8 @@ gst_element_register (GstPlugin *plugin, const gchar *name, guint rank, GType ty
  *
  * Returns: new #GstElement or NULL if the element couldn't be created
  */
-GstElement*
-gst_element_factory_create (GstElementFactory *factory,
-                           const gchar *name)
+GstElement *
+gst_element_factory_create (GstElementFactory * factory, const gchar * name)
 {
   GstElement *element;
   GstElementClass *oclass;
@@ -221,20 +289,27 @@ gst_element_factory_create (GstElementFactory *factory,
     return NULL;
   }
 
-  GST_LOG_OBJECT (factory, "creating element (name \"%s\", type %d)", 
-           GST_STR_NULL (name), (gint) factory->type);
+  if (name)
+    GST_INFO ("creating \"%s\" named \"%s\"", GST_PLUGIN_FEATURE_NAME (factory),
+        GST_STR_NULL (name));
+  else
+    GST_INFO ("creating \"%s\"", GST_PLUGIN_FEATURE_NAME (factory));
 
   if (factory->type == 0) {
-      g_critical ("Factory for `%s' has no type",
-		  GST_PLUGIN_FEATURE_NAME (factory));
-      return NULL;
+    GstPlugin *plugin = GST_PLUGIN_FEATURE (factory)->manager;
+
+    g_critical
+        ("Factory for `%s' has no type. This probably means the plugin wasn't found because the registry is broken. The plugin GStreamer was looking for is named '%s' and is expected in file '%s'. The registry for this plugin is located at '%s'",
+        GST_PLUGIN_FEATURE_NAME (factory),
+        gst_plugin_get_name (plugin), gst_plugin_get_filename (plugin),
+        GST_IS_XML_REGISTRY (plugin->manager) ? GST_XML_REGISTRY (plugin->
+            manager)->location : "Unknown");
+    return NULL;
   }
 
-  oclass = GST_ELEMENT_CLASS (g_type_class_ref (factory->type)); 	 
-  if (oclass->elementfactory == NULL) { 	 
-    GST_DEBUG ("class %s", GST_PLUGIN_FEATURE_NAME (factory)); 	 
+  oclass = GST_ELEMENT_CLASS (g_type_class_ref (factory->type));
+  if (oclass->elementfactory == NULL)
     oclass->elementfactory = factory;
-  }
 
   /* create an instance of the element */
   element = GST_ELEMENT (g_object_new (factory->type, NULL));
@@ -246,6 +321,7 @@ gst_element_factory_create (GstElementFactory *factory,
 
   return element;
 }
+
 /**
  * gst_element_factory_make:
  * @factoryname: a named factory to instantiate
@@ -258,22 +334,21 @@ gst_element_factory_create (GstElementFactory *factory,
  *
  * Returns: new #GstElement or NULL if unable to create element
  */
-GstElement*
-gst_element_factory_make (const gchar *factoryname, const gchar *name)
+GstElement *
+gst_element_factory_make (const gchar * factoryname, const gchar * name)
 {
   GstElementFactory *factory;
   GstElement *element;
 
   g_return_val_if_fail (factoryname != NULL, NULL);
 
-  GST_LOG ("gstelementfactory: make \"%s\" \"%s\"", 
-           factoryname, GST_STR_NULL (name));
+  GST_LOG ("gstelementfactory: make \"%s\" \"%s\"",
+      factoryname, GST_STR_NULL (name));
 
   /* gst_plugin_load_element_factory (factoryname); */
   factory = gst_element_factory_find (factoryname);
   if (factory == NULL) {
-    GST_INFO ("no such element factory \"%s\"!",
-	      factoryname);
+    GST_INFO ("no such element factory \"%s\"!", factoryname);
     return NULL;
   }
   element = gst_element_factory_create (factory, name);
@@ -284,9 +359,10 @@ gst_element_factory_make (const gchar *factoryname, const gchar *name)
 
   return element;
 }
+
 void
-__gst_element_factory_add_pad_template (GstElementFactory *factory,
-					GstPadTemplate *templ)
+__gst_element_factory_add_pad_template (GstElementFactory * factory,
+    GstPadTemplate * templ)
 {
   g_return_if_fail (factory != NULL);
   g_return_if_fail (templ != NULL);
@@ -297,6 +373,7 @@ __gst_element_factory_add_pad_template (GstElementFactory *factory,
   factory->padtemplates = g_list_append (factory->padtemplates, templ);
   factory->numpadtemplates++;
 }
+
 /**
  * gst_element_factory_get_element_type:
  * @factory: factory to get managed #GType from
@@ -306,12 +383,13 @@ __gst_element_factory_add_pad_template (GstElementFactory *factory,
  * Returns: the #GType for elements managed by this factory
  */
 GType
-gst_element_factory_get_element_type (GstElementFactory *factory)
+gst_element_factory_get_element_type (GstElementFactory * factory)
 {
   g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), 0);
 
   return factory->type;
 }
+
 /**
  * gst_element_factory_get_longname:
  * @factory: a #GstElementFactory
@@ -321,12 +399,13 @@ gst_element_factory_get_element_type (GstElementFactory *factory)
  * Returns: the longname
  */
 G_CONST_RETURN gchar *
-gst_element_factory_get_longname (GstElementFactory *factory)
+gst_element_factory_get_longname (GstElementFactory * factory)
 {
   g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), NULL);
 
   return factory->details.longname;
 }
+
 /**
  * gst_element_factory_get_class:
  * @factory: a #GstElementFactory
@@ -336,12 +415,13 @@ gst_element_factory_get_longname (GstElementFactory *factory)
  * Returns: the class
  */
 G_CONST_RETURN gchar *
-gst_element_factory_get_klass (GstElementFactory *factory)
+gst_element_factory_get_klass (GstElementFactory * factory)
 {
   g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), NULL);
 
   return factory->details.klass;
 }
+
 /**
  * gst_element_factory_get_description:
  * @factory: a #GstElementFactory
@@ -351,12 +431,13 @@ gst_element_factory_get_klass (GstElementFactory *factory)
  * Returns: the description
  */
 G_CONST_RETURN gchar *
-gst_element_factory_get_description (GstElementFactory *factory)
+gst_element_factory_get_description (GstElementFactory * factory)
 {
   g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), NULL);
 
   return factory->details.description;
 }
+
 /**
  * gst_element_factory_get_author:
  * @factory: a #GstElementFactory
@@ -366,12 +447,13 @@ gst_element_factory_get_description (GstElementFactory *factory)
  * Returns: the author
  */
 G_CONST_RETURN gchar *
-gst_element_factory_get_author (GstElementFactory *factory)
+gst_element_factory_get_author (GstElementFactory * factory)
 {
   g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), NULL);
 
   return factory->details.author;
 }
+
 /**
  * gst_element_factory_get_num_pad_templates:
  * @factory: a #GstElementFactory
@@ -381,27 +463,83 @@ gst_element_factory_get_author (GstElementFactory *factory)
  * Returns: the number of pad_templates
  */
 guint
-gst_element_factory_get_num_pad_templates (GstElementFactory *factory)
+gst_element_factory_get_num_pad_templates (GstElementFactory * factory)
 {
   g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), 0);
 
   return factory->numpadtemplates;
 }
+
+/**
+ * __gst_element_factory_add_interface:
+ * @elementfactory: The elementfactory to add the interface to
+ * @interfacename: Name of the interface
+ *
+ * Adds the given interfacename to the list of implemented interfaces of the
+ * element.
+ */
+void
+__gst_element_factory_add_interface (GstElementFactory * elementfactory,
+    const gchar * interfacename)
+{
+  g_return_if_fail (GST_IS_ELEMENT_FACTORY (elementfactory));
+  g_return_if_fail (interfacename != NULL);
+  g_return_if_fail (interfacename[0] != '\0');  /* no empty string */
+
+  elementfactory->interfaces =
+      g_list_prepend (elementfactory->interfaces, g_strdup (interfacename));
+}
+
 /**
  * gst_element_factory_get_pad_templates:
  * @factory: a #GstElementFactory
  * 
- * Gets the #Glist of pad templates for this factory.
+ * Gets the #GList of padtemplates for this factory.
  *
  * Returns: the padtemplates
  */
 G_CONST_RETURN GList *
-gst_element_factory_get_pad_templates (GstElementFactory *factory)
+gst_element_factory_get_pad_templates (GstElementFactory * factory)
 {
   g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), NULL);
 
   return factory->padtemplates;
 }
+
+/**
+ * gst_element_factory_get_uri_type:
+ * @factory: a #GstElementFactory
+ * 
+ * Gets the type of URIs the element supports or GST_URI_UNKNOWN if none.
+ *
+ * Returns: type of URIs this element supports
+ */
+guint
+gst_element_factory_get_uri_type (GstElementFactory * factory)
+{
+  g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), GST_URI_UNKNOWN);
+
+  return factory->uri_type;
+}
+
+/**
+ * gst_element_factory_get_uri_protocols:
+ * @factory: a #GstElementFactory
+ * 
+ * Gets a NULL-terminated array of protocols this element supports or NULL, if
+ * no protocols are supported. You may not change the contents of the returned
+ * array as it is still ownt by the element factory. Use g_strdupv() if you want to.
+ *
+ * Returns: the supported protocols or NULL
+ */
+gchar **
+gst_element_factory_get_uri_protocols (GstElementFactory * factory)
+{
+  g_return_val_if_fail (GST_IS_ELEMENT_FACTORY (factory), NULL);
+
+  return factory->uri_protocols;
+}
+
 /**
  * gst_element_factory_can_src_caps :
  * @factory: factory to query
@@ -412,28 +550,30 @@ gst_element_factory_get_pad_templates (GstElementFactory *factory)
  * Returns: true if it can src the capabilities
  */
 gboolean
-gst_element_factory_can_src_caps (GstElementFactory *factory,
-		                 GstCaps *caps)
+gst_element_factory_can_src_caps (GstElementFactory * factory,
+    const GstCaps * caps)
 {
   GList *templates;
 
-  g_return_val_if_fail(factory != NULL, FALSE);
-  g_return_val_if_fail(caps != NULL, FALSE);
+  g_return_val_if_fail (factory != NULL, FALSE);
+  g_return_val_if_fail (caps != NULL, FALSE);
 
   templates = factory->padtemplates;
 
   while (templates) {
-    GstPadTemplate *template = (GstPadTemplate *)templates->data;
+    GstPadTemplate *template = (GstPadTemplate *) templates->data;
 
     if (template->direction == GST_PAD_SRC) {
-      if (gst_caps_is_always_compatible (GST_PAD_TEMPLATE_CAPS (template), caps))
-	return TRUE;
+      if (gst_caps_is_always_compatible (GST_PAD_TEMPLATE_CAPS (template),
+              caps))
+        return TRUE;
     }
     templates = g_list_next (templates);
   }
 
   return FALSE;
 }
+
 /**
  * gst_element_factory_can_sink_caps :
  * @factory: factory to query
@@ -444,22 +584,23 @@ gst_element_factory_can_src_caps (GstElementFactory *factory,
  * Returns: true if it can sink the capabilities
  */
 gboolean
-gst_element_factory_can_sink_caps (GstElementFactory *factory,
-		                  GstCaps *caps)
+gst_element_factory_can_sink_caps (GstElementFactory * factory,
+    const GstCaps * caps)
 {
   GList *templates;
 
-  g_return_val_if_fail(factory != NULL, FALSE);
-  g_return_val_if_fail(caps != NULL, FALSE);
+  g_return_val_if_fail (factory != NULL, FALSE);
+  g_return_val_if_fail (caps != NULL, FALSE);
 
   templates = factory->padtemplates;
 
   while (templates) {
-    GstPadTemplate *template = (GstPadTemplate *)templates->data;
+    GstPadTemplate *template = (GstPadTemplate *) templates->data;
 
     if (template->direction == GST_PAD_SINK) {
-      if (gst_caps_is_always_compatible (caps, GST_PAD_TEMPLATE_CAPS (template)))
-	return TRUE;
+      if (gst_caps_is_always_compatible (caps,
+              GST_PAD_TEMPLATE_CAPS (template)))
+        return TRUE;
     }
     templates = g_list_next (templates);
   }
@@ -467,7 +608,7 @@ gst_element_factory_can_sink_caps (GstElementFactory *factory,
   return FALSE;
 }
 static void
-gst_element_factory_unload_thyself (GstPluginFeature *feature)
+gst_element_factory_unload_thyself (GstPluginFeature * feature)
 {
   GstElementFactory *factory;
 
