@@ -34,8 +34,8 @@ GST_DEBUG_CATEGORY_STATIC (debug_scheduler);
 
 typedef struct _GstSchedulerChain GstSchedulerChain;
 
-#define GST_ELEMENT_THREADSTATE(elem)	(cothread*) (GST_ELEMENT (elem)->sched_private)
-#define GST_RPAD_BUFPEN(pad)		(GstData*)  (GST_REAL_PAD(pad)->sched_private)
+#define GST_ELEMENT_THREADSTATE(elem)	(GST_ELEMENT (elem)->sched_private)
+#define GST_RPAD_BUFPEN(pad)		(GST_REAL_PAD(pad)->sched_private)
 
 #define GST_ELEMENT_COTHREAD_STOPPING			GST_ELEMENT_SCHEDULER_PRIVATE1
 #define GST_ELEMENT_IS_COTHREAD_STOPPING(element)	GST_FLAG_IS_SET((element), GST_ELEMENT_COTHREAD_STOPPING)
@@ -82,14 +82,16 @@ typedef enum
   GST_BASIC_SCHEDULER_STATE_NONE,
   GST_BASIC_SCHEDULER_STATE_STOPPED,
   GST_BASIC_SCHEDULER_STATE_ERROR,
-  GST_BASIC_SCHEDULER_STATE_RUNNING,
-} GstBasicSchedulerState;
+  GST_BASIC_SCHEDULER_STATE_RUNNING
+}
+GstBasicSchedulerState;
 
 typedef enum
 {
   /* something important has changed inside the scheduler */
-  GST_BASIC_SCHEDULER_CHANGE = GST_SCHEDULER_FLAG_LAST,
-} GstBasicSchedulerFlags;
+  GST_BASIC_SCHEDULER_CHANGE = GST_SCHEDULER_FLAG_LAST
+}
+GstBasicSchedulerFlags;
 
 struct _GstBasicScheduler
 {
@@ -127,10 +129,6 @@ static void gst_basic_scheduler_remove_element (GstScheduler * sched,
     GstElement * element);
 static GstElementStateReturn gst_basic_scheduler_state_transition (GstScheduler
     * sched, GstElement * element, gint transition);
-static void gst_basic_scheduler_lock_element (GstScheduler * sched,
-    GstElement * element);
-static void gst_basic_scheduler_unlock_element (GstScheduler * sched,
-    GstElement * element);
 static gboolean gst_basic_scheduler_yield (GstScheduler * sched,
     GstElement * element);
 static gboolean gst_basic_scheduler_interrupt (GstScheduler * sched,
@@ -141,8 +139,8 @@ static void gst_basic_scheduler_pad_link (GstScheduler * sched, GstPad * srcpad,
     GstPad * sinkpad);
 static void gst_basic_scheduler_pad_unlink (GstScheduler * sched,
     GstPad * srcpad, GstPad * sinkpad);
-static void gst_basic_scheduler_pad_select (GstScheduler * sched,
-    GList * padlist);
+static GstData *gst_basic_scheduler_pad_select (GstScheduler * sched,
+    GstPad ** selected, GstPad ** padlist);
 static GstSchedulerState gst_basic_scheduler_iterate (GstScheduler * sched);
 
 static void gst_basic_scheduler_show (GstScheduler * sched);
@@ -167,7 +165,7 @@ static GstSchedulerClass *parent_class = NULL;
   GstElement *current = ((GstBasicScheduler*)sched)->current;	\
   if (current && current->post_run_func)			\
     current->post_run_func (current);				\
-  SCHED (current)->current = NULL;				\
+  ((GstBasicScheduler*) sched)->current = NULL;				\
   do_cothread_switch						\
     (do_cothread_get_main					\
        (((GstBasicScheduler*)sched)->context));			\
@@ -227,10 +225,6 @@ gst_basic_scheduler_class_init (GstBasicSchedulerClass * klass)
       GST_DEBUG_FUNCPTR (gst_basic_scheduler_remove_element);
   gstscheduler_class->state_transition =
       GST_DEBUG_FUNCPTR (gst_basic_scheduler_state_transition);
-  gstscheduler_class->lock_element =
-      GST_DEBUG_FUNCPTR (gst_basic_scheduler_lock_element);
-  gstscheduler_class->unlock_element =
-      GST_DEBUG_FUNCPTR (gst_basic_scheduler_unlock_element);
   gstscheduler_class->yield = GST_DEBUG_FUNCPTR (gst_basic_scheduler_yield);
   gstscheduler_class->interrupt =
       GST_DEBUG_FUNCPTR (gst_basic_scheduler_interrupt);
@@ -256,6 +250,8 @@ gst_basic_scheduler_init (GstBasicScheduler * scheduler)
   scheduler->num_elements = 0;
   scheduler->chains = NULL;
   scheduler->num_chains = 0;
+
+  GST_FLAG_SET (scheduler, GST_SCHEDULER_FLAG_NEW_API);
 }
 
 static void
@@ -269,9 +265,9 @@ plugin_init (GstPlugin * plugin)
 {
   GstSchedulerFactory *factory;
 
-  GST_DEBUG_CATEGORY_INIT (debug_dataflow, "dataflow", 0,
+  GST_DEBUG_CATEGORY_INIT (debug_dataflow, "basic_dataflow", 0,
       "basic scheduler dataflow");
-  GST_DEBUG_CATEGORY_INIT (debug_scheduler, "scheduler", 0,
+  GST_DEBUG_CATEGORY_INIT (debug_scheduler, "basic_scheduler", 0,
       "basic scheduler general information");
 
   factory = gst_scheduler_factory_new ("basic" COTHREADS_NAME,
@@ -366,7 +362,7 @@ gst_basic_scheduler_chain_wrapper (int argc, char **argv)
               GST_CAT_DEBUG (debug_dataflow,
                   "calling chain function of %s:%s %p", name,
                   GST_PAD_NAME (pad), data);
-              GST_RPAD_CHAINFUNC (realpad) (pad, data);
+              gst_pad_call_chain_function (GST_PAD (realpad), data);
               GST_CAT_DEBUG (debug_dataflow,
                   "calling chain function of element %s done", name);
             }
@@ -377,6 +373,11 @@ gst_basic_scheduler_chain_wrapper (int argc, char **argv)
         pads = g_list_next (pads);
       }
     } while (pads != NULL);
+    if (already_iterated == NULL) {
+      GST_DEBUG_OBJECT (SCHED (element), "nothing to iterate for element %s",
+          GST_ELEMENT_NAME (element));
+      break;
+    }
     g_slist_free (already_iterated);
     already_iterated = NULL;
   } while (!GST_ELEMENT_IS_COTHREAD_STOPPING (element));
@@ -392,7 +393,7 @@ gst_basic_scheduler_chain_wrapper (int argc, char **argv)
     SCHED (element)->current = NULL;
   }
 
-  GST_DEBUG ("leaving chain wrapper of element %s", name);
+  GST_CAT_DEBUG (debug_dataflow, "leaving chain wrapper of element %s", name);
   gst_object_unref (GST_OBJECT (element));
 
   return 0;
@@ -405,11 +406,13 @@ gst_basic_scheduler_src_wrapper (int argc, char **argv)
   GList *pads;
   GstRealPad *realpad;
   GstData *data = NULL;
+  gboolean inf_loop;
   G_GNUC_UNUSED const gchar *name = GST_ELEMENT_NAME (element);
 
   GST_DEBUG ("entering src wrapper of element %s", name);
 
   do {
+    inf_loop = TRUE;
     pads = element->pads;
     while (pads) {
 
@@ -419,12 +422,11 @@ gst_basic_scheduler_src_wrapper (int argc, char **argv)
       realpad = GST_REAL_PAD (pads->data);
 
       pads = g_list_next (pads);
-      if (GST_RPAD_DIRECTION (realpad) == GST_PAD_SRC
-          && GST_PAD_IS_USABLE (realpad)) {
+      if (GST_RPAD_DIRECTION (realpad) == GST_PAD_SRC) {
+        inf_loop = FALSE;
         GST_CAT_DEBUG (debug_dataflow, "calling _getfunc for %s:%s",
             GST_DEBUG_PAD_NAME (realpad));
-        g_return_val_if_fail (GST_RPAD_GETFUNC (realpad) != NULL, 0);
-        data = GST_RPAD_GETFUNC (realpad) (GST_PAD (realpad));
+        data = gst_pad_call_get_function (GST_PAD (realpad));
         if (data) {
           GST_CAT_DEBUG (debug_dataflow, "calling gst_pad_push on pad %s:%s %p",
               GST_DEBUG_PAD_NAME (realpad), data);
@@ -432,7 +434,7 @@ gst_basic_scheduler_src_wrapper (int argc, char **argv)
         }
       }
     }
-  } while (!GST_ELEMENT_IS_COTHREAD_STOPPING (element));
+  } while (!GST_ELEMENT_IS_COTHREAD_STOPPING (element) && !inf_loop);
 
   GST_FLAG_UNSET (element, GST_ELEMENT_COTHREAD_STOPPING);
 
@@ -457,7 +459,6 @@ gst_basic_scheduler_chainhandler_proxy (GstPad * pad, GstData * data)
   parent = GST_PAD_PARENT (pad);
   peer = GST_RPAD_PEER (pad);
 
-  GST_DEBUG ("entered chainhandler proxy of %s:%s", GST_DEBUG_PAD_NAME (pad));
   GST_CAT_DEBUG (debug_dataflow, "putting buffer %p in peer \"%s:%s\"'s pen",
       data, GST_DEBUG_PAD_NAME (peer));
 
@@ -494,7 +495,8 @@ gst_basic_scheduler_chainhandler_proxy (GstPad * pad, GstData * data)
 
   do_element_switch (parent);
 
-  GST_DEBUG ("leaving chainhandler proxy of %s:%s", GST_DEBUG_PAD_NAME (pad));
+  GST_CAT_DEBUG (debug_dataflow, "leaving chainhandler proxy of %s:%s",
+      GST_DEBUG_PAD_NAME (pad));
 }
 
 static void
@@ -528,7 +530,8 @@ gst_basic_scheduler_gethandler_proxy (GstPad * pad)
   GstElement *parent;
   GstRealPad *peer;
 
-  GST_DEBUG ("entering gethandler proxy of %s:%s", GST_DEBUG_PAD_NAME (pad));
+  GST_CAT_DEBUG (debug_dataflow, "entering gethandler proxy of %s:%s",
+      GST_DEBUG_PAD_NAME (pad));
 
   parent = GST_PAD_PARENT (pad);
   peer = GST_RPAD_PEER (pad);
@@ -559,7 +562,8 @@ gst_basic_scheduler_gethandler_proxy (GstPad * pad)
   data = GST_RPAD_BUFPEN (pad);
   GST_RPAD_BUFPEN (pad) = NULL;
 
-  GST_DEBUG ("leaving gethandler proxy of %s:%s", GST_DEBUG_PAD_NAME (pad));
+  GST_CAT_DEBUG (debug_dataflow, "leaving gethandler proxy of %s:%s",
+      GST_DEBUG_PAD_NAME (pad));
 
   return data;
 }
@@ -569,7 +573,7 @@ gst_basic_scheduler_eventhandler_proxy (GstPad * srcpad, GstEvent * event)
 {
   gboolean flush;
 
-  GST_INFO ("intercepting event %d on pad %s:%s",
+  GST_CAT_INFO (debug_dataflow, "intercepting event %d on pad %s:%s",
       GST_EVENT_TYPE (event), GST_DEBUG_PAD_NAME (srcpad));
 
   /* figure out if we need to flush */
@@ -589,10 +593,10 @@ gst_basic_scheduler_eventhandler_proxy (GstPad * srcpad, GstEvent * event)
   if (flush) {
     GstData *data = GST_RPAD_BUFPEN (srcpad);
 
-    GST_INFO ("event is flush");
+    GST_CAT_INFO (debug_dataflow, "event is flush");
 
     if (data) {
-      GST_INFO ("need to clear some buffers");
+      GST_CAT_INFO (debug_dataflow, "need to clear some buffers");
 
       gst_data_unref (data);
       GST_RPAD_BUFPEN (srcpad) = NULL;
@@ -691,11 +695,11 @@ gst_basic_scheduler_cothreaded_chain (GstBin * bin, GstSchedulerChain * chain)
             if (GST_RPAD_DIRECTION (peerpad) == GST_PAD_SINK) {
               GST_DEBUG ("copying chain func into push proxy for peer %s:%s",
                   GST_DEBUG_PAD_NAME (peerpad));
-              GST_RPAD_CHAINHANDLER (peerpad) = GST_RPAD_CHAINFUNC (peerpad);
+              GST_RPAD_CHAINHANDLER (peerpad) = gst_pad_call_chain_function;
             } else {
               GST_DEBUG ("copying get func into pull proxy for peer %s:%s",
                   GST_DEBUG_PAD_NAME (peerpad));
-              GST_RPAD_GETHANDLER (peerpad) = GST_RPAD_GETFUNC (peerpad);
+              GST_RPAD_GETHANDLER (peerpad) = gst_pad_call_get_function;
             }
           }
         }
@@ -709,11 +713,11 @@ gst_basic_scheduler_cothreaded_chain (GstBin * bin, GstSchedulerChain * chain)
         if (GST_RPAD_DIRECTION (pad) == GST_PAD_SINK) {
           GST_DEBUG ("copying chain function into push proxy for %s:%s",
               GST_DEBUG_PAD_NAME (pad));
-          GST_RPAD_CHAINHANDLER (pad) = GST_RPAD_CHAINFUNC (pad);
+          GST_RPAD_CHAINHANDLER (pad) = gst_pad_call_chain_function;
         } else {
           GST_DEBUG ("copying get function into pull proxy for %s:%s",
               GST_DEBUG_PAD_NAME (pad));
-          GST_RPAD_GETHANDLER (pad) = GST_RPAD_GETFUNC (pad);
+          GST_RPAD_GETHANDLER (pad) = gst_pad_call_get_function;
         }
       }
       /* otherwise we really are a cothread */
@@ -966,10 +970,11 @@ gst_basic_scheduler_chain_elements (GstBasicScheduler * sched,
     chain1->elements =
         g_list_concat (chain1->elements, g_list_copy (chain2->elements));
     chain1->num_elements += chain2->num_elements;
-    /* FIXME chain changed here */
-/*    gst_basic_scheduler_cothreaded_chain(chain->sched->parent,chain); */
-
     gst_basic_scheduler_chain_destroy (chain2);
+    if (sched->context)
+
+      gst_basic_scheduler_cothreaded_chain (GST_BIN (GST_SCHEDULER (chain1->
+                  sched)->parent), chain1);
 
     /* otherwise one has a chain already, the other doesn't */
   } else {
@@ -1193,20 +1198,6 @@ gst_basic_scheduler_state_transition (GstScheduler * sched,
   return GST_STATE_SUCCESS;
 }
 
-static void
-gst_basic_scheduler_lock_element (GstScheduler * sched, GstElement * element)
-{
-  if (GST_ELEMENT_THREADSTATE (element))
-    do_cothread_lock (GST_ELEMENT_THREADSTATE (element));
-}
-
-static void
-gst_basic_scheduler_unlock_element (GstScheduler * sched, GstElement * element)
-{
-  if (GST_ELEMENT_THREADSTATE (element))
-    do_cothread_unlock (GST_ELEMENT_THREADSTATE (element));
-}
-
 static gboolean
 gst_basic_scheduler_yield (GstScheduler * sched, GstElement * element)
 {
@@ -1264,11 +1255,7 @@ gst_basic_scheduler_pad_link (GstScheduler * sched, GstPad * srcpad,
   GST_DEBUG ("srcpad sched is %p, sinkpad sched is %p",
       GST_ELEMENT_SCHED (srcelement), GST_ELEMENT_SCHED (sinkelement));
 
-  if (GST_ELEMENT_SCHED (srcelement) == GST_ELEMENT_SCHED (sinkelement)) {
-    GST_INFO ("peer %s:%s is in same scheduler, chaining together",
-        GST_DEBUG_PAD_NAME (sinkpad));
-    gst_basic_scheduler_chain_elements (bsched, srcelement, sinkelement);
-  }
+  gst_basic_scheduler_chain_elements (bsched, srcelement, sinkelement);
 }
 
 static void
@@ -1332,40 +1319,40 @@ gst_basic_scheduler_pad_unlink (GstScheduler * sched, GstPad * srcpad,
 #endif
 }
 
-static void
-gst_basic_scheduler_pad_select (GstScheduler * sched, GList * padlist)
+static GstData *
+gst_basic_scheduler_pad_select (GstScheduler * sched, GstPad ** selected,
+    GstPad ** padlist)
 {
-  GstPad *pad = NULL;
-  GList *padlist2 = padlist;
+  GstData *data = NULL;
+  gint i = 0;
 
   GST_INFO ("performing select");
 
-  while (padlist2) {
-    pad = GST_PAD (padlist2->data);
-
-
-    padlist2 = g_list_next (padlist2);
-  }
-
-  /* else there is nothing ready to consume, set up the select functions */
-  while (padlist) {
-    pad = GST_PAD (padlist->data);
+  while (padlist[i]) {
+    GstPad *pad = padlist[i];
 
     GST_RPAD_CHAINHANDLER (pad) =
         GST_DEBUG_FUNCPTR (gst_basic_scheduler_select_proxy);
-
-    padlist = g_list_next (padlist);
   }
-  if (pad != NULL) {
-    GstRealPad *peer = GST_RPAD_PEER (pad);
 
-    do_element_switch (GST_PAD_PARENT (peer));
+  do_element_switch (GST_PAD_PARENT (GST_PAD_PEER (padlist[0])));
 
-    /* FIXME disabled for now */
-    /* pad = GST_ELEMENT (GST_PAD_PARENT (pad))->select_pad; */
+  i = 0;
+  while (padlist[i]) {
+    GstPad *pad = padlist[i];
 
-    g_assert (pad != NULL);
+    if (GST_RPAD_BUFPEN (pad)) {
+      *selected = pad;
+      data = GST_RPAD_BUFPEN (pad);
+      GST_RPAD_BUFPEN (pad) = NULL;
+    }
+
+    GST_RPAD_CHAINHANDLER (pad) =
+        GST_DEBUG_FUNCPTR (gst_basic_scheduler_chainhandler_proxy);
   }
+
+  g_assert (data != NULL);
+  return data;
 }
 
 static GstSchedulerState
