@@ -39,13 +39,15 @@ GST_DEBUG_CATEGORY_STATIC (debug_scheduler);
 #define GST_ELEMENT_SCHED_CONTEXT(elem)		((GstOptSchedulerCtx*) (GST_ELEMENT (elem)->sched_private))
 #define GST_ELEMENT_SCHED_GROUP(elem)		(GST_ELEMENT_SCHED_CONTEXT (elem)->group)
 /* need this first macro to not run into lvalue casts */
-#define GST_PAD_BUFPEN(pad)			(GST_REAL_PAD(pad)->sched_private)
-#define GST_PAD_BUFLIST(pad)            	((GList*) GST_PAD_BUFPEN(pad))
+#define GST_PAD_DATAPEN(pad)			(GST_REAL_PAD(pad)->sched_private)
+#define GST_PAD_DATALIST(pad)            	((GList*) GST_PAD_DATAPEN(pad))
 
 #define GST_ELEMENT_COTHREAD_STOPPING			GST_ELEMENT_SCHEDULER_PRIVATE1
 #define GST_ELEMENT_IS_COTHREAD_STOPPING(element)	GST_FLAG_IS_SET((element), GST_ELEMENT_COTHREAD_STOPPING)
-#define GST_ELEMENT_INTERRUPTED				GST_ELEMENT_SCHEDULER_PRIVATE2
-#define GST_ELEMENT_IS_INTERRUPTED(element)		GST_FLAG_IS_SET((element), GST_ELEMENT_INTERRUPTED)
+#define GST_ELEMENT_VISITED				GST_ELEMENT_SCHEDULER_PRIVATE2
+#define GST_ELEMENT_IS_VISITED(element)			GST_FLAG_IS_SET((element), GST_ELEMENT_VISITED)
+#define GST_ELEMENT_SET_VISITED(element)		GST_FLAG_SET((element), GST_ELEMENT_VISITED)
+#define GST_ELEMENT_UNSET_VISITED(element)		GST_FLAG_UNSET((element), GST_ELEMENT_VISITED)
 
 typedef struct _GstOptScheduler GstOptScheduler;
 typedef struct _GstOptSchedulerClass GstOptSchedulerClass;
@@ -89,6 +91,9 @@ struct _GstOptScheduler
   gint recursion;
 
   gint max_recursion;
+  gint live_groups;
+  gint live_chains;
+  gint live_links;
 };
 
 struct _GstOptSchedulerClass
@@ -183,6 +188,7 @@ struct _GstOptSchedulerGroup
   GstOptSchedulerChain *chain;  /* the chain this group belongs to */
   GstOptSchedulerGroupFlags flags;      /* flags for this group */
   GstOptSchedulerGroupType type;        /* flags for this group */
+  GstOptScheduler *sched;       /* the scheduler */
 
   gint refcount;
 
@@ -202,7 +208,6 @@ struct _GstOptSchedulerGroup
   char **argv;
 };
 
-
 /* 
  * A group is a set of elements through which data can flow without switching
  * cothreads or without invoking the scheduler's run queue.
@@ -229,6 +234,7 @@ static void group_error_handler (GstOptSchedulerGroup * group);
 static void group_element_set_enabled (GstOptSchedulerGroup * group,
     GstElement * element, gboolean enabled);
 static gboolean schedule_group (GstOptSchedulerGroup * group);
+static void get_group (GstElement * element, GstOptSchedulerGroup ** group);
 
 
 /* 
@@ -427,6 +433,9 @@ gst_opt_scheduler_init (GstOptScheduler * scheduler)
   scheduler->elements = NULL;
   scheduler->iterations = 1;
   scheduler->max_recursion = 100;
+  scheduler->live_groups = 0;
+  scheduler->live_chains = 0;
+  scheduler->live_links = 0;
 }
 
 static void
@@ -499,11 +508,12 @@ create_chain (GstOptScheduler * osched)
   chain->sched = osched;
   chain->refcount = 1;
   chain->flags = GST_OPT_SCHEDULER_CHAIN_DISABLED;
+  osched->live_chains++;
 
   gst_object_ref (GST_OBJECT (osched));
   osched->chains = g_slist_prepend (osched->chains, chain);
 
-  GST_LOG ("new chain %p", chain);
+  GST_LOG ("new chain %p, %d live chains now", chain, osched->live_chains);
 
   return chain;
 }
@@ -520,6 +530,9 @@ destroy_chain (GstOptSchedulerChain * chain)
 
   osched = chain->sched;
   osched->chains = g_slist_remove (osched->chains, chain);
+  osched->live_chains--;
+
+  GST_LOG ("%d live chains now", osched->live_chains);
 
   gst_object_unref (GST_OBJECT (osched));
 
@@ -651,7 +664,7 @@ merge_chains (GstOptSchedulerChain * chain1, GstOptSchedulerChain * chain2)
 }
 
 /* sorts the group list so that terminal sinks come first -- prevents pileup of
- * buffers in bufpens */
+ * datas in datapens */
 static void
 sort_chain (GstOptSchedulerChain * chain)
 {
@@ -777,12 +790,9 @@ chain_recursively_migrate_group (GstOptSchedulerChain * chain,
   add_to_chain (chain, group);
 
   /* then follow all links */
-  links = group->group_links;
-  while (links) {
+  for (links = group->group_links; links; links = g_slist_next (links)) {
     GstOptSchedulerGroupLink *link = (GstOptSchedulerGroupLink *) links->data;
 
-    links = g_slist_next (links);
-
     chain_recursively_migrate_group (chain, OTHER_GROUP_LINK (link, group));
   }
 }
@@ -822,11 +832,14 @@ create_group (GstOptSchedulerChain * chain, GstElement * element,
   group->refcount = 1;          /* float... */
   group->flags = GST_OPT_SCHEDULER_GROUP_DISABLED;
   group->type = type;
+  group->sched = chain->sched;
+  group->sched->live_groups++;
 
   add_to_group (group, element, FALSE);
   add_to_chain (chain, group);
   group = unref_group (group);  /* ...and sink. */
 
+  GST_LOG ("%d live groups now", group->sched->live_groups);
   /* group's refcount is now 2 (one for the element, one for the chain) */
 
   return group;
@@ -845,6 +858,9 @@ destroy_group (GstOptSchedulerGroup * group)
   if (group->flags & GST_OPT_SCHEDULER_GROUP_SCHEDULABLE)
     destroy_group_scheduler (group);
 
+  group->sched->live_groups--;
+  GST_LOG ("%d live groups now", group->sched->live_groups);
+
   g_free (group);
 }
 
@@ -927,6 +943,37 @@ remove_from_group (GstOptSchedulerGroup * group, GstElement * element)
   return group;
 }
 
+/* count number of elements in the group. Have to be careful because
+ * decoupled elements are added as entry point but are not added to
+ * the elements list */
+static gint
+group_num_elements (GstOptSchedulerGroup * group)
+{
+  gint num;
+
+  num = group->num_elements;
+  /* decoupled elements are not added to the group but are
+   * added as an entry */
+  if (group->entry) {
+    if (GST_ELEMENT_IS_DECOUPLED (group->entry)) {
+      num++;
+    }
+  }
+  return num;
+}
+
+/* check if an element is part of the given group. We have to be carefull
+ * as decoupled elements are added as entry but are not added to the elements 
+ * list */
+static gboolean
+group_has_element (GstOptSchedulerGroup * group, GstElement *element)
+{
+  if (group->entry == element)
+    return TRUE;
+
+  return (g_slist_find (group->elements, element) != NULL);
+}
+
 /* FIXME need to check if the groups are of the same type -- otherwise need to
    setup the scheduler again, if it is setup */
 static GstOptSchedulerGroup *
@@ -1087,7 +1134,10 @@ schedule_group (GstOptSchedulerGroup * group)
 {
   if (!group->entry) {
     GST_INFO ("not scheduling group %p without entry", group);
-    return FALSE;
+    /* FIXME, we return true here, while the group is actually
+     * not schedulable. We might want to disable the element that caused
+     * this group to be scheduled instead */
+    return TRUE;
   }
 #ifdef USE_COTHREADS
   if (group->cothread)
@@ -1102,23 +1152,28 @@ schedule_group (GstOptSchedulerGroup * group)
     GST_INFO ("not scheduling group %p without schedulefunc", group);
     return FALSE;
   } else {
-    GSList *l;
+    GSList *l, *lcopy;
 
-    for (l = group->elements; l; l = l->next) {
+    lcopy = g_slist_copy (group->elements);
+    for (l = lcopy; l; l = l->next) {
       GstElement *e = (GstElement *) l->data;
 
+      gst_object_ref (GST_OBJECT (e));
       if (e->pre_run_func)
         e->pre_run_func (e);
     }
 
     group->schedulefunc (group->argc, group->argv);
 
-    for (l = group->elements; l; l = l->next) {
+    for (l = lcopy; l; l = l->next) {
       GstElement *e = (GstElement *) l->data;
 
       if (e->post_run_func)
         e->post_run_func (e);
+
+      gst_object_unref (GST_OBJECT (e));
     }
+    g_slist_free (lcopy);
 
   }
   return TRUE;
@@ -1154,7 +1209,13 @@ gst_opt_scheduler_schedule_run_queue (GstOptScheduler * osched)
 
     GST_LOG_OBJECT (osched, "scheduling group %p", group);
 
-    res = schedule_group (group);
+    if (GST_OPT_SCHEDULER_GROUP_IS_ENABLED (group)) {
+      res = schedule_group (group);
+    } else {
+      GST_INFO_OBJECT (osched,
+           "group was disabled while it was on the queue, not scheduling");
+      res = TRUE;
+    }
     if (!res) {
       g_warning ("error scheduling group %p", group);
       group_error_handler (group);
@@ -1245,6 +1306,7 @@ get_group_schedule_function (int argc, char *argv[])
     data = gst_pad_call_get_function (pad);
     if (data) {
       if (GST_EVENT_IS_INTERRUPT (data)) {
+        GST_DEBUG ("unreffing interrupt event %p", data);
         gst_event_unref (GST_EVENT (data));
         break;
       }
@@ -1279,7 +1341,7 @@ loop_group_schedule_function (int argc, char *argv[])
   else
     group_error_handler (group);
 
-  GST_LOG ("loopfunc ended of element %s in group %p",
+  GST_LOG ("returned from loopfunc of element %s in group %p",
       GST_ELEMENT_NAME (entry), group);
 
   group->flags &= ~GST_OPT_SCHEDULER_GROUP_RUNNING;
@@ -1303,7 +1365,7 @@ unknown_group_schedule_function (int argc, char *argv[])
 
 /* this function is called when the first element of a chain-loop or a loop-loop
  * link performs a push to the loop element. We then schedule the
- * group with the loop-based element until the bufpen is empty */
+ * group with the loop-based element until the datapen is empty */
 static void
 gst_opt_scheduler_loop_wrapper (GstPad * sinkpad, GstData * data)
 {
@@ -1318,19 +1380,19 @@ gst_opt_scheduler_loop_wrapper (GstPad * sinkpad, GstData * data)
   GST_LOG ("chain handler for loop-based pad %" GST_PTR_FORMAT, sinkpad);
 
 #ifdef USE_COTHREADS
-  if (GST_PAD_BUFLIST (peer)) {
+  if (GST_PAD_DATALIST (peer)) {
     g_warning ("deadlock detected, disabling group %p", group);
     group_error_handler (group);
   } else {
-    GST_LOG ("queueing data %p on %s:%s's bufpen", data,
+    GST_LOG ("queueing data %p on %s:%s's datapen", data,
         GST_DEBUG_PAD_NAME (peer));
-    GST_PAD_BUFPEN (peer) = g_list_append (GST_PAD_BUFLIST (peer), data);
+    GST_PAD_DATAPEN (peer) = g_list_append (GST_PAD_DATALIST (peer), data);
     schedule_group (group);
   }
 #else
-  GST_LOG ("queueing data %p on %s:%s's bufpen", data,
+  GST_LOG ("queueing data %p on %s:%s's datapen", data,
       GST_DEBUG_PAD_NAME (peer));
-  GST_PAD_BUFPEN (peer) = g_list_append (GST_PAD_BUFLIST (peer), data);
+  GST_PAD_DATAPEN (peer) = g_list_append (GST_PAD_DATALIST (peer), data);
   if (!(group->flags & GST_OPT_SCHEDULER_GROUP_RUNNING)) {
     GST_LOG ("adding group %p to runqueue", group);
     if (!g_list_find (osched->runqueue, group)) {
@@ -1340,13 +1402,13 @@ gst_opt_scheduler_loop_wrapper (GstPad * sinkpad, GstData * data)
   }
 #endif
 
-  GST_LOG ("%d buffers left on %s:%s's bufpen after chain handler",
-      g_list_length (GST_PAD_BUFLIST (peer)));
+  GST_LOG ("%d datas left on %s:%s's datapen after chain handler",
+      g_list_length (GST_PAD_DATALIST (peer)), GST_DEBUG_PAD_NAME (peer));
 }
 
 /* this function is called by a loop based element that performs a
- * pull on a sinkpad. We schedule the peer group until the bufpen
- * is filled with the buffer so that this function  can return */
+ * pull on a sinkpad. We schedule the peer group until the datapen
+ * is filled with the data so that this function can return */
 static GstData *
 gst_opt_scheduler_get_wrapper (GstPad * srcpad)
 {
@@ -1357,24 +1419,31 @@ gst_opt_scheduler_get_wrapper (GstPad * srcpad)
 
   GST_LOG ("get handler for %" GST_PTR_FORMAT, srcpad);
 
-  /* first try to grab a queued buffer */
-  if (GST_PAD_BUFLIST (srcpad)) {
-    data = GST_PAD_BUFLIST (srcpad)->data;
-    GST_PAD_BUFPEN (srcpad) = g_list_remove (GST_PAD_BUFLIST (srcpad), data);
+  /* first try to grab a queued data */
+  if (GST_PAD_DATALIST (srcpad)) {
+    data = GST_PAD_DATALIST (srcpad)->data;
+    GST_PAD_DATAPEN (srcpad) = g_list_remove (GST_PAD_DATALIST (srcpad), data);
 
     GST_LOG ("returning popped queued data %p", data);
 
     return data;
   }
+  GST_LOG ("need to schedule the peer element");
 
   /* else we need to schedule the peer element */
-  group = GST_ELEMENT_SCHED_GROUP (GST_PAD_PARENT (srcpad));
+  get_group (GST_PAD_PARENT (srcpad), &group);
+  if (group == NULL) {
+    /* wow, peer has no group */
+    GST_LOG ("peer without group detected");
+    //group_error_handler (group);
+    return GST_DATA (gst_event_new (GST_EVENT_INTERRUPT));
+  }
   osched = group->chain->sched;
   data = NULL;
   disabled = FALSE;
 
   do {
-    GST_LOG ("scheduling upstream group %p to fill bufpen", group);
+    GST_LOG ("scheduling upstream group %p to fill datapen", group);
 #ifdef USE_COTHREADS
     schedule_group (group);
 #else
@@ -1406,18 +1475,18 @@ gst_opt_scheduler_get_wrapper (GstPad * srcpad)
       return GST_DATA (gst_event_new (GST_EVENT_INTERRUPT));
     }
 #endif
-    /* if the scheduler interrupted, make sure we send an INTERRUPTED event to the
-     * loop based element */
+    /* if the scheduler interrupted, make sure we send an INTERRUPTED event
+     * to the loop based element */
     if (osched->state == GST_OPT_SCHEDULER_STATE_INTERRUPTED) {
       GST_INFO ("scheduler interrupted, return interrupt event");
       data = GST_DATA (gst_event_new (GST_EVENT_INTERRUPT));
     } else {
-      if (GST_PAD_BUFLIST (srcpad)) {
-        data = GST_PAD_BUFLIST (srcpad)->data;
-        GST_PAD_BUFPEN (srcpad) =
-            g_list_remove (GST_PAD_BUFLIST (srcpad), data);
+      if (GST_PAD_DATALIST (srcpad)) {
+        data = GST_PAD_DATALIST (srcpad)->data;
+        GST_PAD_DATAPEN (srcpad) =
+            g_list_remove (GST_PAD_DATALIST (srcpad), data);
       } else if (disabled) {
-        /* no buffer in queue and peer group was disabled */
+        /* no data in queue and peer group was disabled */
         data = GST_DATA (gst_event_new (GST_EVENT_INTERRUPT));
       }
     }
@@ -1425,7 +1494,7 @@ gst_opt_scheduler_get_wrapper (GstPad * srcpad)
   while (data == NULL);
 
   GST_LOG ("get handler, returning data %p, queue length %d",
-      data, g_list_length (GST_PAD_BUFLIST (srcpad)));
+      data, g_list_length (GST_PAD_DATALIST (srcpad)));
 
   return data;
 }
@@ -1433,13 +1502,13 @@ gst_opt_scheduler_get_wrapper (GstPad * srcpad)
 static void
 pad_clear_queued (GstPad * srcpad, gpointer user_data)
 {
-  GList *buflist = GST_PAD_BUFLIST (srcpad);
+  GList *datalist = GST_PAD_DATALIST (srcpad);
 
-  if (buflist) {
-    GST_LOG ("need to clear some buffers");
-    g_list_foreach (buflist, (GFunc) gst_data_unref, NULL);
-    g_list_free (buflist);
-    GST_PAD_BUFPEN (srcpad) = NULL;
+  if (datalist) {
+    GST_LOG ("need to clear some datas");
+    g_list_foreach (datalist, (GFunc) gst_data_unref, NULL);
+    g_list_free (datalist);
+    GST_PAD_DATAPEN (srcpad) = NULL;
   }
 }
 
@@ -1448,7 +1517,7 @@ gst_opt_scheduler_event_wrapper (GstPad * srcpad, GstEvent * event)
 {
   gboolean flush;
 
-  GST_DEBUG ("intercepting event %d on pad %s:%s",
+  GST_DEBUG ("intercepting event type %d on pad %s:%s",
       GST_EVENT_TYPE (event), GST_DEBUG_PAD_NAME (srcpad));
 
   /* figure out if this is a flush event */
@@ -1466,7 +1535,7 @@ gst_opt_scheduler_event_wrapper (GstPad * srcpad, GstEvent * event)
   }
 
   if (flush) {
-    GST_LOG ("event is flush");
+    GST_LOG ("event triggers a flush");
 
     pad_clear_queued (srcpad, NULL);
   }
@@ -1480,7 +1549,7 @@ gst_opt_scheduler_state_transition (GstScheduler * sched, GstElement * element,
   GstOptSchedulerGroup *group;
   GstElementStateReturn res = GST_STATE_SUCCESS;
 
-  GST_DEBUG ("element \"%s\" state change %d", GST_ELEMENT_NAME (element),
+  GST_DEBUG ("element \"%s\" state change (%04x)", GST_ELEMENT_NAME (element),
       transition);
 
   /* we check the state of the managing pipeline here */
@@ -1553,8 +1622,6 @@ get_group (GstElement * element, GstOptSchedulerGroup ** group)
 {
   GstOptSchedulerCtx *ctx;
 
-  /*GList *pads; */
-
   ctx = GST_ELEMENT_SCHED_CONTEXT (element);
   if (ctx)
     *group = ctx->group;
@@ -1654,7 +1721,10 @@ group_inc_link (GstOptSchedulerGroup * src, GstOptSchedulerGroup * sink)
     src->group_links = g_slist_prepend (src->group_links, link);
     sink->group_links = g_slist_prepend (sink->group_links, link);
 
-    GST_DEBUG ("added group link between %p and %p", src, sink);
+    src->sched->live_links++;
+
+    GST_DEBUG ("added group link between %p and %p, %d live links now",
+        src, sink, src->sched->live_links);
   }
 }
 
@@ -1680,10 +1750,33 @@ group_dec_link (GstOptSchedulerGroup * group1, GstOptSchedulerGroup * group2)
       GST_LOG ("link count between %p and %p is now %d",
           group1, group2, link->count);
       if (link->count == 0) {
+        GstOptSchedulerGroup *iso_group = NULL;
+
         group1->group_links = g_slist_remove (group1->group_links, link);
         group2->group_links = g_slist_remove (group2->group_links, link);
+        group1->sched->live_links--;
+
+        GST_LOG ("%d live links now", group1->sched->live_links);
+
         g_free (link);
         GST_DEBUG ("removed group link between %p and %p", group1, group2);
+        if (group1->group_links == NULL) {
+          /* group1 has no more links with other groups */
+          iso_group = group1;
+        } else if (group2->group_links == NULL) {
+          /* group2 has no more links with other groups */
+          iso_group = group2;
+        }
+        if (iso_group) {
+          GstOptSchedulerChain *chain;
+
+          GST_DEBUG ("group %p has become isolated, moving to new chain",
+              iso_group);
+
+          chain = create_chain (iso_group->chain->sched);
+          remove_from_chain (iso_group->chain, iso_group);
+          add_to_chain (chain, iso_group);
+        }
         res = TRUE;
       }
       break;
@@ -1970,6 +2063,8 @@ gst_opt_scheduler_pad_link (GstScheduler * sched, GstPad * srcpad,
 
         GST_DEBUG ("setting \"%s\" as entry point of _get-based group %p",
             GST_ELEMENT_NAME (src_element), group);
+
+        setup_group_scheduler (osched, group);
       }
       break;
     }
@@ -2021,7 +2116,8 @@ gst_opt_scheduler_pad_link (GstScheduler * sched, GstPad * srcpad,
        * group1 only exists if src_element is linked to some other element */
       if (!group1) {
         /* create a new group for src_element as it cannot be merged into another group
-         * here. we create the group in the same chain as the loop-based element. */
+         * here. we create the group in the same chain as the loop-based element. note
+         * that creating a new group will also increment the links with other groups */
         GST_DEBUG ("creating new group for element %s",
             GST_ELEMENT_NAME (src_element));
         group1 =
@@ -2031,8 +2127,9 @@ gst_opt_scheduler_pad_link (GstScheduler * sched, GstPad * srcpad,
         /* both elements are already in a group, make sure they are added to
          * the same chain */
         merge_chains (group1->chain, group2->chain);
+        /* increment the group link counters */
+        group_inc_link (group1, group2);
       }
-      group_inc_link (group1, group2);
       break;
     }
     case GST_OPT_INVALID:
@@ -2041,21 +2138,42 @@ gst_opt_scheduler_pad_link (GstScheduler * sched, GstPad * srcpad,
   }
 }
 
-/* 
- * checks if an element is still linked to some other element in the group. 
- * no checking is done on the brokenpad arg 
- */
-static gboolean
-element_has_link_with_group (GstElement * element, GstOptSchedulerGroup * group,
+static void
+group_elements_set_visited (GstOptSchedulerGroup * group, gboolean visited)
+{
+  GSList *elements;
+
+  for (elements = group->elements; elements; elements = g_slist_next (elements)) {
+    GstElement *element = GST_ELEMENT (elements->data);
+
+    if (visited) {
+      GST_ELEMENT_SET_VISITED (element);
+    } else {
+      GST_ELEMENT_UNSET_VISITED (element);
+    }
+  }
+}
+
+static GList *
+element_get_reachables_func (GstElement * element, GstOptSchedulerGroup * group,
     GstPad * brokenpad)
 {
-  gboolean linked = FALSE;
+  GList *result = NULL;
   const GList *pads;
 
-  /* see if the element has no more links to the peer group */
+  /* if no element or element not in group or been there, return NULL */
+  if (element == NULL || !group_has_element (group, element) ||
+      GST_ELEMENT_IS_VISITED (element))
+    return NULL;
+
+  GST_ELEMENT_SET_VISITED (element);
+
+  result = g_list_prepend (result, element);
+
   pads = gst_element_get_pad_list (element);
-  while (pads && !linked) {
+  while (pads) {
     GstPad *pad = GST_PAD (pads->data);
+    GstPad *peer;
 
     pads = g_list_next (pads);
 
@@ -2063,27 +2181,40 @@ element_has_link_with_group (GstElement * element, GstOptSchedulerGroup * group,
     if (!GST_IS_REAL_PAD (pad) || pad == brokenpad)
       continue;
 
-    if (GST_PAD_PEER (pad)) {
+    peer = GST_PAD_PEER (pad);
+    if (!GST_IS_REAL_PAD (peer) || peer == brokenpad)
+      continue;
+
+    if (peer) {
       GstElement *parent;
-      GstOptSchedulerGroup *parentgroup;
+      GList *res;
 
-      /* see in what group this element is */
-      parent = GST_PAD_PARENT (GST_PAD_PEER (pad));
+      parent = GST_PAD_PARENT (peer);
 
-      /* links with decoupled elements are valid */
-      if (GST_ELEMENT_IS_DECOUPLED (parent)) {
-        linked = TRUE;
-      } else {
-        /* for non-decoupled elements we need to check the group */
-        get_group (parent, &parentgroup);
+      res = element_get_reachables_func (parent, group, brokenpad);
 
-        /* if it's in the same group, we're still linked */
-        if (parentgroup == group)
-          linked = TRUE;
-      }
+      result = g_list_concat (result, res);
     }
   }
-  return linked;
+
+  return result;
+}
+
+static GList *
+element_get_reachables (GstElement * element, GstOptSchedulerGroup * group,
+    GstPad * brokenpad)
+{
+  GList *result;
+
+  /* reset visited flags */
+  group_elements_set_visited (group, FALSE);
+
+  result = element_get_reachables_func (element, group, brokenpad);
+
+  /* and reset visited flags again */
+  group_elements_set_visited (group, FALSE);
+
+  return result;
 }
 
 /* 
@@ -2126,7 +2257,7 @@ group_can_reach_group (GstOptSchedulerGroup * group,
 
     GST_LOG ("found link from %p to %p, count %d", group, other, link->count);
 
-    /* check if we can reach the target recursiveley */
+    /* check if we can reach the target recursively */
     reachable = group_can_reach_group (other, target);
   }
   /* unset the visited flag, note that this is not optimal as we might be checking
@@ -2174,16 +2305,154 @@ group_inc_links_for_element (GstOptSchedulerGroup * group, GstElement * element)
   GstPad *pad;
   GstOptSchedulerGroup *peer_group;
 
+  GST_DEBUG ("group %p, element %s ", group, gst_element_get_name (element));
+
   for (l = GST_ELEMENT_PADS (element); l; l = l->next) {
     pad = (GstPad *) l->data;
     if (GST_IS_REAL_PAD (pad) && GST_PAD_PEER (pad)) {
       get_group (GST_PAD_PARENT (GST_PAD_PEER (pad)), &peer_group);
       if (peer_group && peer_group != group)
-        group_inc_link (group, peer_group);
+        if (GST_PAD_DIRECTION (pad) == GST_PAD_SRC)
+          group_inc_link (group, peer_group);
+        else
+          group_inc_link (peer_group, group);
     }
   }
 }
 
+static void
+debug_element (GstElement * element, GstOptScheduler * osched)
+{
+  GST_LOG ("element %s", gst_element_get_name (element));
+}
+
+/* move this group in the chain of the groups it has links with */
+static void
+rechain_group (GstOptSchedulerGroup * group)
+{
+  GstOptSchedulerChain *chain = NULL;
+  GSList *links;
+
+  GST_LOG ("checking if this group needs rechaining");
+
+  /* follow all links */
+  for (links = group->group_links; links; links = g_slist_next (links)) {
+    GstOptSchedulerGroupLink *link = (GstOptSchedulerGroupLink *) links->data;
+    GstOptSchedulerGroup *other;
+
+    other = OTHER_GROUP_LINK (link, group);
+    GST_LOG ("found link with other group %p with chain %p", other,
+        other->chain);
+
+    /* first time, take chain */
+    if (chain == NULL) {
+      chain = other->chain;
+    }
+    /* second time, chain should be the same */
+    else if (other->chain != chain) {
+      g_warning ("(internal error): chain inconsistency");
+    }
+  }
+  if (!chain) {
+    GST_LOG ("no new chain found, not rechaining");
+  } else if (chain != group->chain) {
+    GST_LOG ("need to move group %p to chain %p", group, chain);
+    /* remove from old chain */
+    remove_from_chain (group->chain, group);
+    /* and move to new chain */
+    add_to_chain (chain, group);
+  } else {
+    GST_LOG ("group %p is in correct chain %p", group, chain);
+  }
+}
+
+/* migrate the element and all connected elements to a new group without looking at
+ * the brokenpad */
+static GstOptSchedulerGroup *
+group_migrate_connected (GstOptScheduler * osched, GstElement * element,
+    GstOptSchedulerGroup * group, GstPad * brokenpad)
+{
+  GList *connected, *c;
+  GstOptSchedulerGroup *new_group = NULL, *tst;
+  GstOptSchedulerChain *chain;
+  gint len;
+
+  if (GST_ELEMENT_IS_DECOUPLED (element)) {
+    GST_LOG ("element is decoupled and thus not in the group");
+    /* the element is decoupled and is therefore not in the group */
+    return NULL;
+  }
+
+  get_group (element, &tst);
+  if (tst == NULL) {
+    GST_LOG ("element has no group, not interesting");
+    return NULL;
+  }
+
+  GST_LOG ("migrate connected elements to new group");
+  connected = element_get_reachables (element, group, brokenpad);
+  GST_LOG ("elements to move to new group:");
+  g_list_foreach (connected, (GFunc) debug_element, NULL);
+
+  len = g_list_length (connected);
+
+  if (len == 0) {
+    g_warning ("(internal error) found lost element %s",
+        gst_element_get_name (element));
+    return NULL;
+  } else if (len == 1) {
+    group = remove_from_group (group, GST_ELEMENT (connected->data));
+    GST_LOG
+        ("not migrating to new group as the group would only contain 1 element");
+    g_list_free (connected);
+    GST_LOG ("new group is old group now");
+    new_group = group;
+  } else {
+    /* we create a new chain to hold the new group */
+    chain = create_chain (osched);
+
+    for (c = connected; c; c = g_list_next (c)) {
+      GstElement *element = GST_ELEMENT (c->data);
+
+      group = remove_from_group (group, element);
+      if (new_group == NULL) {
+        new_group =
+            create_group (chain, element, GST_OPT_SCHEDULER_GROUP_UNKNOWN);
+      } else {
+        add_to_group (new_group, element, TRUE);
+      }
+    }
+    g_list_free (connected);
+
+    /* remove last element from the group if any. Make sure not to remove
+     * the loop based entry point of a group as this always needs one group */
+    if (group != NULL) {
+      if (group_num_elements (group) == 1 &&
+          group->type != GST_OPT_SCHEDULER_GROUP_LOOP) {
+        GST_LOG ("removing last element from old group");
+        group = remove_from_group (group, GST_ELEMENT (group->elements->data));
+      }
+    }
+  }
+
+  if (new_group != NULL) {
+    if (group_num_elements (new_group) == 1 &&
+        new_group->type != GST_OPT_SCHEDULER_GROUP_LOOP) {
+      GST_LOG ("removing last element from new group");
+      new_group =
+          remove_from_group (new_group, GST_ELEMENT (new_group->elements->data));
+      return NULL;
+    }
+    /* at this point the new group lives in its own chain but might
+     * have to be merged with another chain, this happens when the new
+     * group has a link with another group in another chain */
+    rechain_group (new_group);
+  }
+
+
+  return new_group;
+}
+
 static void
 gst_opt_scheduler_pad_unlink (GstScheduler * sched,
     GstPad * srcpad, GstPad * sinkpad)
@@ -2252,7 +2521,6 @@ gst_opt_scheduler_pad_unlink (GstScheduler * sched,
   }
   /* hard part, groups are equal */
   else {
-    gboolean still_link1, still_link2;
     GstOptSchedulerGroup *group;
 
     /* since group1 == group2, it doesn't matter which group we take */
@@ -2260,97 +2528,45 @@ gst_opt_scheduler_pad_unlink (GstScheduler * sched,
 
     GST_LOG ("elements are in the same group %p", group);
 
-    /* check if the element is still linked to some other element in the group,
-     * we pass the pad that is broken up as an arg because a link on that pad
-     * is not valid anymore.
-     * Note that this check is only to make sure that a single element can be removed 
-     * completely from the group, we also have to check for migrating several 
-     * elements to a new group. */
-    still_link1 = element_has_link_with_group (src_element, group, srcpad);
-    still_link2 = element_has_link_with_group (sink_element, group, sinkpad);
-    /* if there is still a link, we don't need to break this group */
-    if (still_link1 && still_link2) {
-      GSList *l;
-      GList *m;
-      int linkcount;
+    if (group->entry == NULL) {
+      /* it doesn't really matter, we just have to make sure that both 
+       * elements end up in another group if they are not connected */
+      GST_LOG ("group %p has no entry, moving source element to new group",
+          group);
+      group_migrate_connected (osched, src_element, group, srcpad);
+    } else {
+      GList *reachables;
 
-      GST_LOG ("elements still have links with other elements in the group");
+      GST_LOG ("group %p has entry %p", group, group->entry);
 
-      while (group->elements)
-        for (l = group->elements; l && l->data; l = l->next) {
-          GstElement *element = (GstElement *) l->data;
+      /* get of a list of all elements that are still managed by the old
+       * entry element */
+      reachables = element_get_reachables (group->entry, group, srcpad);
+      GST_LOG ("elements still reachable from the entry:");
+      g_list_foreach (reachables, (GFunc) debug_element, sched);
 
-          if (!element || !GST_IS_ELEMENT (element) ||
-              GST_ELEMENT_IS_DECOUPLED (element))
-            continue;
-
-          linkcount = 0;
-          GST_LOG ("Examining %s\n", GST_ELEMENT_NAME (element));
-          for (m = GST_ELEMENT_PADS (element); m; m = m->next) {
-            GstPad *peer, *pad;
-            GstElement *parent;
-            GstOptSchedulerGroup *peer_group;
-
-            pad = (GstPad *) m->data;
-            if (!pad || !GST_IS_REAL_PAD (pad))
-              continue;
-
-            peer = GST_PAD_PEER (pad);
-            if (!peer || !GST_IS_REAL_PAD (peer))
-              continue;
-
-            parent = GST_PAD_PARENT (GST_PAD_PEER (pad));
-            get_group (parent, &peer_group);
-            if (peer_group && peer_group != group) {
-              GST_LOG ("pad %s is linked with %s\n",
-                  GST_PAD_NAME (pad), GST_ELEMENT_NAME (parent));
-              linkcount++;
-            }
-          }
-
-          if (linkcount < 2) {
-            remove_from_group (group, element);
-          }
-          /* if linkcount == 2, it will be unlinked later on */
-          else if (linkcount > 2) {
-            g_warning
-                ("opt: Can't handle element %s with 3 or more links, aborting",
-                GST_ELEMENT_NAME (element));
-            return;
-          }
-        }
-      /* Peer element will be caught during next iteration */
-      return;
-    }
-
-    /* now check which one of the elements we can remove from the group */
-    if (!still_link1) {
-      /* we only remove elements that are not the entry point of a loop based
-       * group and are not decoupled */
-      if (!(group->entry == src_element &&
-              group->type == GST_OPT_SCHEDULER_GROUP_LOOP) &&
-          !GST_ELEMENT_IS_DECOUPLED (src_element)) {
-        GST_LOG ("el ement1 is separated from the group");
-
-        remove_from_group (group, src_element);
+      /* if the source is reachable from the entry, we can leave it in the group */
+      if (g_list_find (reachables, src_element)) {
+        GST_LOG
+            ("source element still reachable from the entry, leaving in group");
       } else {
-        GST_LOG ("src_element is decoupled or entry in loop based group");
+        GST_LOG
+            ("source element not reachable from the entry, moving to new group");
+        group_migrate_connected (osched, src_element, group, srcpad);
       }
-    }
 
-    if (!still_link2) {
-      /* we only remove elements that are not the entry point of a loop based
-       * group and are not decoupled */
-      if (!(group->entry == sink_element &&
-              group->type == GST_OPT_SCHEDULER_GROUP_LOOP) &&
-          !GST_ELEMENT_IS_DECOUPLED (sink_element)) {
-        GST_LOG ("sink_element is separated from the group");
-
-        remove_from_group (group, sink_element);
+      /* if the sink is reachable from the entry, we can leave it in the group */
+      if (g_list_find (reachables, sink_element)) {
+        GST_LOG
+            ("sink element still reachable from the entry, leaving in group");
       } else {
-        GST_LOG ("sink_element is decoupled or entry in loop based group");
+        GST_LOG
+            ("sink element not reachable from the entry, moving to new group");
+        group_migrate_connected (osched, sink_element, group, srcpad);
       }
+      g_list_free (reachables);
     }
+    /* at this point the group can be freed and gone, so don't touch */
   }
 }
 
@@ -2364,8 +2580,6 @@ gst_opt_scheduler_iterate (GstScheduler * sched)
 
   osched->state = GST_OPT_SCHEDULER_STATE_RUNNING;
 
-  //gst_opt_scheduler_show (sched);
-
   GST_DEBUG_OBJECT (sched, "iterating");
 
   while (iterations) {
@@ -2383,6 +2597,7 @@ gst_opt_scheduler_iterate (GstScheduler * sched)
         GST_LOG ("scheduling chain %p", chain);
         schedule_chain (chain);
         scheduled = TRUE;
+        GST_LOG ("scheduled chain %p", chain);
       } else {
         GST_LOG ("not scheduling disabled chain %p", chain);
       }
